|
|     Amiga Linux/68k Loader
|
|      Copyright 1995 by Geert Uytterhoeven (Geert.Uytterhoeven@cs.kuleuven.ac.be)
|
|     -----------------------------------------------------------------------------
|
|     The original idea for this program comes from the Linux distribution:
|
|        LILO  -  Generic Boot Loader for Linux ("LInux LOader")
|                 by Werner Almesberger
|
|     -----------------------------------------------------------------------------
|
|     This file is subject to the terms and conditions of the GNU General Public
|     License.  See the file README.legal in the main directory of the Linux/68k
|     distribution for more details.
|


.text
.globl _BootCodeStart, _BootCodeEnd, _LoaderVectorStart, _LoaderVectorEnd
.align 4


|	Alert Codes

AT_DeadEnd		= 0x80000000
AN_LILO			= AT_DeadEnd+0x4c000000
AG_NoMemory		= 0x00010000
AG_IOError		= 0x00060000
AG_BadParm		= 0x00080000
NOT_ENOUGH_MEMORY	= AN_LILO+AG_NoMemory	| Yields Alert Code CC010000
READ_ERROR		= AN_LILO+AG_IOError	| Yields Alert Code CC060000
HUNK_ERROR		= AN_LILO+AG_BadParm	| Yields Alert Code CC080000

|	Memory Allocation

MEMF_PUBLIC		= 1
MEMF_CLEAR		= 0x10000
MEMF_PUB_CLR		= MEMF_PUBLIC+MEMF_CLEAR

|	Exec I/O

IO_COMMAND		= 28
IO_FLAGS		= 30
IO_LENGTH		= 36
IO_DATA			= 40
IO_OFFSET		= 44
CMD_READ		= 2
IOF_QUICK		= 1

|	Trackdisk

TD_SECTOR		= 512
TD_SECSHIFT		= 9
TD_MOTOR		= 9

|	Resident Tag

RT_INIT			= 22

|	Expansion Library

eb_Flags		= 34
EBB_SILENTSTART		= 6

|	DOS Hunks

HUNK_CODE		= 1001
HUNK_DATA		= 1002
HUNK_BSS		= 1003
HUNK_RELOC32		= 1004
HUNK_END		= 1010
HUNK_HEADER		= 1011

|	Exec Library Vector Offsets

_LVOFindResident	= -0x60
_LVOAlert		= -0x6c
_LVOCloseLibrary	= -0x19e
_LVODoIO		= -0x1c8
_LVOOpenLibrary		= -0x228
_LVOCopyMem		= -0x270
_LVOCacheClearU		= -0x27c
_LVOAllocVec		= -0x2ac
_LVOFreeVec		= -0x2b2


_BootCodeStart:

|
|	Boot Code
|
|		in:	a1 = bootrequest
|			a6 = execbase
|
|		out:	d0 = error
|

Entry:			lea	(BASE.w,pc),a5
			move.l	a1,(BootRequest-BASE.w,a5)
			move.l	a6,(SysBase-BASE.w,a5)

|	Set Up the Function Table

			lea	(FuncTable.w,pc),a0
			lea	(lilo_open.w,pc),a1
			move.l	a1,(a0)+
			lea	(lilo_close.w,pc),a1
			move.l	a1,(a0)+
			lea	(lilo_read.w,pc),a1
			move.l	a1,(a0)+
			lea	(lilo_loadseg.w,pc),a1
			move.l	a1,(a0)+
			lea	(lilo_unloadseg.w,pc),a1
			move.l	a1,(a0)+
			lea	(lilo_exec.w,pc),a1
			move.l	a1,(a0)+

|	Allocate the File Buffer

			move.l	#TD_SECTOR,d0
			move.l	#MEMF_PUB_CLR,d1
			jsr	(_LVOAllocVec,a6)
			move.l	d0,(FileBuffer-BASE.w,a5)
			jne	BufferOK
			jra	NoMemory

|	Load and Execute the Boot Loader

BufferOK:		lea	(_LoaderVectorStart.w,pc),a0
			jbsr	lilo_loadseg
			move.l	d0,a2
			lea	(FuncTable.w,pc),a0
			move.l	a0,d0
			move.l	a2,a0
			jbsr	lilo_exec
			move.l	d0,d2
			move.l	a2,a0
			jbsr	lilo_unloadseg
			move.l	(FileBuffer.w,pc),a1
			jsr	(_LVOFreeVec,a6)
			tst.l	d2
			jne	BootAmigaOS
BootError:		moveq	#-1,d0
			rts

|	Boot the Standard Amiga Operating System

BootAmigaOS:		lea	(ExpansionName.w,pc),a1
			moveq	#37,d0
			jsr	(_LVOOpenLibrary,a6)
			tst.l	d0
			jeq	NoExpansion37
			move.l	d0,a1
			bset	#EBB_SILENTSTART,(eb_Flags,a1)
			jsr	(_LVOCloseLibrary,a6)
NoExpansion37:		lea	(DOSName.w,pc),a1
			jsr	(_LVOFindResident,a6)
			tst.l	d0
			jeq	BootError
			move.l	d0,a0
			move.l	(RT_INIT,a0),a0
			moveq	#0,d0
			rts


|
|	Open a File
|
|		in:	a0 = vector
|
|		out:	void
|

lilo_open:		move.l	a5,-(a7)
			lea	(BASE.w,pc),a5
			move.l	a0,(FileVector-BASE.w,a5)
			clr.l	(FileOffset-BASE.w,a5)
			move.l	(a7)+,a5
			rts

|
|	Close a File
|
|		in:	void
|
|		out:	void
|

lilo_close:		move.l	a6,-(a7)
			move.l	(BootRequest.w,pc),a1
			move.w	#TD_MOTOR,(IO_COMMAND,a1)
			move.b	#IOF_QUICK,(IO_FLAGS,a1)
			clr.l	(IO_LENGTH,a1)
			move.l	(SysBase.w,pc),a6
			jsr	(_LVODoIO,a6)
			move.l	(a7)+,a6
			rts

|
|	Read from the File
|
|		in:	a0 = dest
|			d0 = length
|
|		out:	void
|

lilo_read:		movem.l	d2-d3/a2/a5-a6,-(a7)
			lea	(BASE.w,pc),a5
			move.l	(SysBase.w,pc),a6
			move.l	d0,d2
			move.l	a0,a2
ReadLoop:		tst.l	d2
			jeq	End
			move.l	(FileOffset.w,pc),d3
			move.l	d3,d0
			and.l	#TD_SECTOR-1,d3
			jne	NoLoad
			bchg	#1,(0xbfe001)
			moveq	#TD_SECSHIFT,d1
			asr.l	d1,d0
			move.l	(FileVector.w,pc),a0
			move.l	(4,a0,d0.l*4),a0
			move.l	(BootRequest.w,pc),a1
			move.w	#CMD_READ,(IO_COMMAND,a1)
			move.b	#IOF_QUICK,(IO_FLAGS,a1)
			move.l	#TD_SECTOR,(IO_LENGTH,a1)
			move.l	(FileBuffer.w,pc),(IO_DATA,a1)
			move.l	a0,(IO_OFFSET,a1)
			jsr	(_LVODoIO,a6)
			tst.l	d0
			jeq	NoLoad
			move.l	#READ_ERROR,d7
			jmp	(_LVOAlert,a6)
NoLoad:			move.l	#TD_SECTOR,d0
			sub.l	d3,d0
			cmp.l	d2,d0
			jlt	NotAll
			move.l	d2,d0
NotAll:			move.l	(FileBuffer.w,pc),a0
			lea	(a0,d3.l),a0
			move.l	a2,a1
			move.l	d0,d3
			jsr	(_LVOCopyMem,a6)
			sub.l	d3,d2
			add.l	d3,a2
			add.l	d3,(FileOffset-BASE.w,a5)
			jra	ReadLoop
End:			movem.l	(a7)+,d2-d3/a2/a5-a6
			rts

|
|	Read a LONG from the File
|
|		in:	void
|
|		out:	d0 = data
|

ReadLong:		subq.l	#4,a7
			lea	(a7),a0
			moveq	#4,d0
			jbsr	lilo_read
			move.l	(a7)+,d0
			rts

|
|	Load an AmigaOS Executable
|
|		in:	a0 = vector
|
|		out:	d0 = table
|

lilo_loadseg:		movem.l	d2-d5/a2-a3/a6,-(a7)
			move.l	(SysBase.w,pc),a6
			jbsr	lilo_open
			jbsr	ReadLong
			cmp.l	#HUNK_HEADER,d0
			jeq	HeaderOK
HunkError:		move.l	#HUNK_ERROR,d7
			jmp	(_LVOAlert,a6)
HeaderOK:		jbsr	ReadLong
			tst.l	d0
			jne	HunkError
			jbsr	ReadLong
			move.l	d0,d5
			addq.l	#1,d0
			asl.l	#2,d0
			move.l	#MEMF_PUB_CLR,d1
			jsr	(_LVOAllocVec,a6)
			tst.l	d0
			jne	TableOK
NoMemory:		move.l	#NOT_ENOUGH_MEMORY,d7
			jmp	(_LVOAlert,a6)
TableOK:		move.l	d0,a2
			move.l	d5,(a2)
			jbsr	ReadLong
			tst.l	d0
			jne	HunkError
			jbsr	ReadLong
			addq.l	#1,d0
			cmp.l	d0,d5
			jne	HunkError
			moveq	#0,d2
AllocLoop:		jbsr	ReadLong
			asl.l	#2,d0
			move.l	#MEMF_PUB_CLR,d1
			jsr	(_LVOAllocVec,a6)
			move.l	d0,(4,a2,d2.l*4)
			jeq	NoMemory
			addq.l	#1,d2
			cmp.l	d2,d5
			jne	AllocLoop
			moveq	#0,d2
NextHunk:		move.l	(4,a2,d2.l*4),a3
			jbsr	ReadLong
			cmp.l	#HUNK_END,d0
			jeq	HunkEnd
			cmp.l	#HUNK_CODE,d0
			jeq	HunkCodeData
			cmp.l	#HUNK_DATA,d0
			jeq	HunkCodeData
			cmp.l	#HUNK_BSS,d0
			jeq	HunkBss
			cmp.l	#HUNK_RELOC32,d0
			jne	HunkError
HunkReloc32:		jbsr	ReadLong
			move.l	d0,d3
			jeq	NextHunk
			jbsr	ReadLong
			move.l	(4,a2,d0.l*4),d4
			subq.l	#1,d3
RelocLoop:		jbsr	ReadLong
			add.l	d4,(a3,d0.l)
			dbf	d3,RelocLoop
			jra	HunkReloc32
HunkCodeData:		jbsr	ReadLong
			move.l	a3,a0
			asl.l	#2,d0
			jbsr	lilo_read
			jra	NextHunk
HunkBss:		jbsr	ReadLong
			jra	NextHunk
HunkEnd:		addq.l	#1,d2
			cmp.l	d2,d5
			jne	NextHunk
			jbsr	lilo_close
			move.l	a2,d0
			movem.l	(a7)+,d2-d5/a2-a3/a6
			rts


|
|	Unload an AmigaOS Executable
|
|		in:	a0 = table
|
|		out:	void
|

lilo_unloadseg:		movem.l	d2/a2-a3/a6,-(a7)
			move.l	a0,a2
			move.l	a0,a3
			move.l	(a2)+,d2
			subq.l	#1,d2
			move.l	(SysBase.w,pc),a6
FreeLoop:		move.l	(a2)+,a1
			jsr	(_LVOFreeVec,a6)
			dbf	d2,FreeLoop
			move.l	a3,a1
			jsr	(_LVOFreeVec,a6)
			movem.l	(a7)+,d2/a2-a3/a6
			rts


|
|	Execute an AmigaOS Executable
|
|		in:	a0 = table
|			d0 = args
|
|		out:	d0 = result
|

lilo_exec:		movem.l	d2/a2/a6,-(a7)
			move.l	(4,a0),a2
			move.l	d0,d2
			move.l	(SysBase.w,pc),a6
			jsr	(_LVOCacheClearU,a6)
			move.l	d2,d0
			jsr	(a2)
			movem.l	(a7)+,d2/a2/a6
			rts


|
|	Lilo Data
|

			.align 4

BASE:


|
|	ExecBase
|

SysBase:		.long	0


|
|	Function Table
|

FuncTable:		.long	0			| lilo_open
			.long	0			| lilo_close
			.long	0			| lilo_read
			.long	0			| lilo_loadseg
			.long	0			| lilo_unloadseg
			.long	0			| lilo_exec

|
|	Boot IORequest passed by Bootstrap
|

BootRequest:		.long	0


|
|	File Stuff
|

FileVector:		.long	0
FileOffset:		.long	0
FileBuffer:		.long	0


|
|	File Data for the Loader Program
|

_LoaderVectorStart:	.long	0			| File Size
			.long	0,0,0,0,0,0,0,0,0,0,0,0	| Data Blocks
			.long	0,0,0,0,0,0,0,0,0,0,0,0
			.long	0,0,0,0,0,0,0,0,0,0,0,0
			.long	0,0,0,0,0,0,0,0,0,0,0,0
			.long	0,0,0,0,0,0,0,0
_LoaderVectorEnd:


|
|	Strings
|

ExpansionName:		.ascii	"expansion.library\0"
DOSName:		.ascii	"dos.library\0"

_BootCodeEnd:
