/*
 *		Amiga Linux/68k Loader
 *
 *		 Copyright 1995 by Geert Uytterhoeven (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *		-----------------------------------------------------------------------------
 *
 *		The original idea for this program comes from the Linux distribution:
 *
 *			LILO  -  Generic Boot Loader for Linux ("LInux LOader")
 *						by Werner Almesberger
 *
 *		-----------------------------------------------------------------------------
 *
 *		This file is subject to the terms and conditions of the GNU General Public
 *		License.  See the file README.legal in the main directory of the Linux/68k
 *		distribution for more details.
 */


#define LILO_ID					(0x4C494C4F)			/* 'LILO' */
#define LILO_MAPVECTORID		(0x4D415056)			/* 'MAPV' */

#define AMIGASECTORSIZE			(512)

#define LILO_CONFIGFILE			"/etc/lilo.conf"
#define LILO_MAPFILE				"/boot/map"
#define LILO_LOADERFILE			"/boot/loader.patched"
#define LILO_LOADERTEMPLATE	"/boot/loader"


#define ID_BOOT_DISK				(0x424F4F55)	/* 'BOOU' */
#define ID_OFS_DISK				(0x444F5300)	/* 'DOS\0' */
#define ID_FFS_DISK				(0x444F5301)	/* 'DOS\1' */
#define ID_OFS_INTL_DISK		(0x444F5302)	/* 'DOS\2' */
#define ID_FFS_INTL_DISK		(0x444F5303)	/* 'DOS\3' */
#define ID_OFS_DC_DISK			(0x444F5304)	/* 'DOS\4' */
#define ID_FFS_DC_DISK			(0x444F5305)	/* 'DOS\5' */
#define ID_MUFS_DISK				(0x6d754653)	/* 'muFS' */
#define ID_MU_OFS_DISK			(0x6d754600)	/* 'muF\0' */
#define ID_MU_FFS_DISK			(0x6d754601)	/* 'muF\1' */
#define ID_MU_OFS_INTL_DISK	(0x6d754602)	/* 'muF\2' */
#define ID_MU_FFS_INTL_DISK	(0x6d754603)	/* 'muF\3' */
#define ID_MU_OFS_DC_DISK		(0x6d754604)	/* 'muF\4' */
#define ID_MU_FFS_DC_DISK		(0x6d754605)	/* 'muF\5' */


struct BootBlock {
	u_long ID;
	u_long Checksum;
	u_long DosBlock;
	u_long Data[252];
	u_long LiloID;
};


	/*
	 *		Boot Operating Systems
	 */

#define BOS_AMIGA		(0)					/* AmigaOS (always available: -0 == 0 :-) */
#define BOS_LINUX		(1)					/* Linux/68k */
#define BOS_NETBSD	(2)					/* Future Expansion :-) */


	/*
	 *		Map File
	 */

#define MAXLABELSIZE	32
#define MAXPWDSIZE	32


struct MapHeader {
	u_long LiloID;
	u_long Auto;								/* Automatic Boot */
	u_long Aux;									/* Future Expansion: Use Serial Console */
	u_long TimeOut;							/* In Seconds, 0 is Infinite */
	u_char MasterPwd[MAXPWDSIZE];	
};


struct MapEntry {
	union {
		u_long NotLast;						/* On Disk */
		struct MapEntry *Next;				/* In Memory */
	} un;
	u_long Size;
	u_char Label[MAXLABELSIZE];
	u_char Password[MAXPWDSIZE];
	long Type;									/* Negative for an Unavailable Boot Image */
	u_char CommandLine[CL_SIZE];
	u_long KernelVector[0];
};
