/*
 *		Amiga Linux/68k Loader
 *
 *		 Copyright 1995 by Geert Uytterhoeven (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *		-----------------------------------------------------------------------------
 *
 *		This program is based on files of the Linux/68k distribution:
 *
 *		 o bootstrap.c -- This program loads the Linux/68k kernel into an Amiga
 *								and launches it.
 *
 *			Copyright 1993, 1994 by Hamish Macdonald, Greg Harp
 *
 *			Modified 11-May-94 by Geert Uytterhoeven: A3640 MapROM check
 *			Modified 31-May-94 by Geert Uytterhoeven: Memory thrash problem solved
 *
 *		 o bootstrap.h -- This file is a part of the Amiga bootloader.
 *
 *			Copyright 1993, 1994 by Hamish Macdonald
 *
 *			Some minor additions by Michael Rausch 1-11-94
 *			Modified 11-May-94 by Geert Uytterhoeven: inline Supervisor() call
 *
 *
 *		The original idea for this program comes from the Linux distribution:
 *
 *			LILO  -  Generic Boot Loader for Linux ("LInux LOader")
 *						by Werner Almesberger
 *
 *		-----------------------------------------------------------------------------
 *
 *		This file is subject to the terms and conditions of the GNU General Public
 *		License.  See the file README.legal in the main directory of the Linux/68k
 *		distribution for more details.
 */


	/*
	 *		Exec Library Definitions
	 */

#define TRUE	(1)
#define FALSE	(0)

#define AT_DeadEnd		0x80000000
#define AG_NoMemory		0x00010000
#define AG_NoSignal		0x00070000
#define AG_OpenLib		0x00030000
#define AG_OpenDev		0x00040000
#define AG_OpenRes		0x00050000
#define AO_GraphicsLib	0x00008002
#define AO_Intuition		0x00008004
#define AO_ExpansionLib	0x0000800A
#define AO_ConsoleDev	0x00008011
#define AO_TimerDev		0x00008015
#define AO_Unknown		0x00008035

struct MsgPort {
	u_char fill1[15];
	u_char mp_SigBit;
	u_char fill2[18];
};

struct IOStdReq {
	u_char fill1[20];
	struct Device *io_Device;
	u_char fill2[4];
	u_short io_Command;
	u_char io_Flags;
	u_char fill3[5];
	u_long io_Length;
	void *io_Data;
	u_char fill4[4];
};

#define CMD_READ			2
#define CMD_WRITE			3
#define IOF_QUICK			(1<<0)

struct timerequest {
	u_char fill1[28];
	u_short io_Command;
	u_char io_Flags;
	u_char fill2[1];
	u_long tv_secs;
	u_long tv_micro;
};

#define UNIT_VBLANK		1
#define TR_ADDREQUEST	9

struct List {
	struct Node *lh_Head;
	struct Node *lh_Tail;
	struct Node *lh_TailPred;
	u_char lh_Type;
	u_char l_pad;
};

struct MemChunk {
	struct MemChunk *mc_Next;	/* pointer to next chunk */
	u_long mc_Bytes;				/* chunk byte size	*/
};

#define MEMF_PUBLIC	(1<<0)
#define MEMF_CHIP		(1<<1)
#define MEMF_FAST		(1<<2)
#define MEMF_LOCAL	(1<<8)
#define MEMF_CLEAR	(1<<16)

struct MemHeader {
	struct Node mh_Node;
	u_short mh_Attributes;		/* characteristics of this region */
	struct MemChunk *mh_First;	/* first free region		*/
	void *mh_Lower;				/* lower memory bound		*/
	void *mh_Upper;				/* upper memory bound+1 */
	u_long mh_Free;				/* total number of free bytes	*/
};

struct ExecBase {
	u_char fill1[296];
	u_short AttnFlags;
	u_char fill2[24];
	struct List MemList;
	u_char fill3[194];
	u_char VBlankFrequency;
	u_char PowerSupplyFrequency;
	u_char fill4[36];
	u_long ex_EClockFrequency;
	u_char fill5[60];
};

#define AFB_68020		1
#define AFF_68020		(1<<AFB_68020)
#define AFB_68030		2
#define AFF_68030		(1<<AFB_68030)
#define AFB_68040		3
#define AFF_68040		(1<<AFB_68040)
#define AFB_68881		4
#define AFF_68881		(1<<AFB_68881)
#define AFB_68882		5
#define AFF_68882		(1<<AFB_68882)
#define AFB_FPU40		6
#define AFF_FPU40		(1<<AFB_FPU40)

#define CONU_SNIPMAP			3
#define CONFLAG_DEFAULT		0

#define IEQUALIFIER_LSHIFT				0x0001
#define IEQUALIFIER_RSHIFT				0x0002
#define IEQUALIFIER_CAPSLOCK			0x0004
#define IEQUALIFIER_CONTROL			0x0008
#define IEQUALIFIER_LALT				0x0010
#define IEQUALIFIER_RALT				0x0020
#define IEQUALIFIER_LCOMMAND			0x0040
#define IEQUALIFIER_RCOMMAND			0x0080
#define IEQUALIFIER_MIDBUTTON			0x1000
#define IEQUALIFIER_RBUTTON			0x2000
#define IEQUALIFIER_LEFTBUTTON		0x4000

struct IORequest;
struct Library;


	/*
	 *		Exec Library Functions
	 */

extern struct ExecBase *SysBase;

static __inline unsigned long Supervisor(unsigned long (*userfunc)())
{
	register unsigned long _res __asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register unsigned long (*a0)() __asm("a0") = userfunc;
		/* gcc doesn't seem to like asm parameters in a5 */
	__asm __volatile ("movel a5,-(sp);"
							"movel a0,a5;"
							"jsr (-0x1e,a6);"
							"movel (sp)+,a5"
							: "=r" (_res)
							: "r" (a6), "r" (a0)
							: "a0","a1","d0","d1","memory");
	return(_res);
}

static __inline void Alert(unsigned long alertNum)
{
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register unsigned long d7 __asm("d7") = alertNum;

	__asm __volatile ("jsr (-0x6c,a6)"
							: /* no output */
							: "r" (a6), "r" (d7)
							: "a0","a1","d0","d1", "memory");
}

static __inline void Disable(void)
{
	register struct ExecBase *a6 __asm("a6") = SysBase;

	__asm __volatile ("jsr (-0x78,a6)"
							: /* no output */
							: "r" (a6)
							: "a0","a1","d0","d1", "memory");
}

static __inline void *SuperState(void)
{
	register void  *_res	__asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;

	__asm __volatile ("jsr (-0x96,a6)"
							: "=r" (_res)
							: "r" (a6)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline void *AllocMem(unsigned long byteSize, unsigned long requirements)
{
	register void  *_res	__asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register unsigned long d0 __asm("d0") = byteSize;
	register unsigned long d1 __asm("d1") = requirements;

	__asm __volatile ("jsr (-0xc6,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (d0), "r" (d1)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline void FreeMem(void * memoryBlock, unsigned long byteSize)
{
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register void *a1 __asm("a1") = memoryBlock;
	register unsigned long d0 __asm("d0") = byteSize;

	__asm __volatile ("jsr (-0xd2,a6)"
							: /* no output */
							: "r" (a6), "r" (a1), "r" (d0)
							: "a0","a1","d0","d1", "memory");
}

static __inline u_long Wait(unsigned long signalSet)
{
	register u_long _res __asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register unsigned long d0 __asm("d0") = signalSet;

	__asm __volatile ("jsr (-0x13e,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (d0)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline void CloseLibrary(struct Library *library)
{
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register struct Library *a1 __asm("a1") = library;

	__asm __volatile ("jsr (-0x19e,a6)"
							: /* no output */
							: "r" (a6), "r" (a1)
							: "a0","a1","d0","d1", "memory");
}

static __inline char OpenDevice(u_char *devName, unsigned long unit, struct IORequest *ioRequest, unsigned long flags)
{
	register char _res __asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register u_char *a0 __asm("a0") = devName;
	register unsigned long d0 __asm("d0") = unit;
	register struct IORequest *a1 __asm("a1") = ioRequest;
	register unsigned long d1 __asm("d1") = flags;

	__asm __volatile ("jsr (-0x1bc,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (a0), "r" (a1), "r" (d0), "r" (d1)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline void CloseDevice(struct IORequest *ioRequest)
{
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register struct IORequest *a1 __asm("a1") = ioRequest;

	__asm __volatile ("jsr (-0x1c2,a6)"
							: /* no output */
							: "r" (a6), "r" (a1)
							: "a0","a1","d0","d1", "memory");
}

static __inline char DoIO(struct IORequest *ioRequest)
{
	register char _res __asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register struct IORequest *a1 __asm("a1") = ioRequest;

	__asm __volatile ("jsr (-0x1c8,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (a1)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline void SendIO(struct IORequest *ioRequest)
{
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register struct IORequest *a1 __asm("a1") = ioRequest;

	__asm __volatile ("jsr (-0x1ce,a6)"
							: /* no output */
							: "r" (a6), "r" (a1)
							: "a0","a1","d0","d1", "memory");
}

static __inline char WaitIO(struct IORequest *ioRequest)
{
	register char _res __asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register struct IORequest *a1 __asm("a1") = ioRequest;

	__asm __volatile ("jsr (-0x1da,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (a1)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline char AbortIO(struct IORequest *ioRequest)
{
	register char _res __asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register struct IORequest *a1 __asm("a1") = ioRequest;

	__asm __volatile ("jsr (-0x1e0,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (a1)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline void *RawDoFmt(u_char *formatString, void *dataStream, void (*putChProc)(), void *putChData)
{
	register void *_res __asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register u_char *a0 __asm("a0") = formatString;
	register void *a1 __asm("a1") = dataStream;
	register void (*a2)() __asm("a2") = putChProc;
	register void *a3 __asm("a3") = putChData;

	__asm __volatile ("jsr (-0x20a,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (a0), "r" (a1), "r" (a2), "r" (a3)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline struct Library *OpenLibrary(u_char *libName, unsigned long version)
{
	register struct Library *_res __asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register u_char *a1 __asm("a1") = libName;
	register unsigned long d0 __asm("d0") = version;

	__asm __volatile ("jsr (-0x228,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (a1), "r" (d0)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline unsigned long CacheControl(unsigned long cacheBits, unsigned long cacheMask)
{
	register unsigned long  _res  __asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register unsigned long d0 __asm("d0") = cacheBits;
	register unsigned long d1 __asm("d1") = cacheMask;

	__asm __volatile ("jsr (-0x288,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (d0), "r" (d1)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline void *CreateIORequest(struct MsgPort *port, unsigned long size)
{
	register struct Library *_res __asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register struct MsgPort *a0 __asm("a0") = port;
	register unsigned long d0 __asm("d0") = size;

	__asm __volatile ("jsr (-0x28e,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (a0), "r" (d0)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline void DeleteIORequest(void *ioRequest)
{
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register void *a0 __asm("a0") = ioRequest;

	__asm __volatile ("jsr (-0x294,a6)"
							: /* no output */
							: "r" (a6), "r" (a0)
							: "a0","a1","d0","d1", "memory");
}

static __inline struct MsgPort *CreateMsgPort(void)
{
	register struct MsgPort *_res __asm("d0");
	register struct ExecBase *a6 __asm("a6") = SysBase;

	__asm __volatile ("jsr (-0x29a,a6)"
							: "=r" (_res)
							: "r" (a6)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline void DeleteMsgPort(struct MsgPort *port)
{
	register struct ExecBase *a6 __asm("a6") = SysBase;
	register struct MsgPort *a0 __asm("a0") = port;

	__asm __volatile ("jsr (-0x2a0,a6)"
							: /* no output */
							: "r" (a6), "r" (a0)
							: "a0","a1","d0","d1", "memory");
}


	/*
	 *		Expansion Library Functions
	 */

extern struct ExpansionBase *ExpansionBase;

static __inline struct ConfigDev *FindConfigDev(struct ConfigDev *oldConfigDev, long manufacturer, long product)
{
	register struct ConfigDev * _res  __asm("d0");
	register struct ExpansionBase* a6 __asm("a6") = ExpansionBase;
	register struct ConfigDev *a0 __asm("a0") = oldConfigDev;
	register long d0 __asm("d0") = manufacturer;
	register long d1 __asm("d1") = product;

	__asm __volatile ("jsr (-0x48,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (a0), "r" (d0), "r" (d1)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}


	/*
	 *		Graphics Library Definitions
	 */

struct GfxBase {
	u_char Fill1[20];
	u_short Version;
	u_char Fill2[194];
	u_short NormalDisplayRows;
	u_short NormalDisplayColumns;
	u_char Fill3[16];
	u_char ChipRevBits0;
	u_char Fill4[307];
};

#define GFXB_HR_AGNUS	0
#define GFXF_HR_AGNUS	(1<<GFXB_HR_AGNUS)
#define GFXB_HR_DENISE	1
#define GFXF_HR_DENISE	(1<<GFXB_HR_DENISE)
#define GFXB_AA_ALICE	2
#define GFXF_AA_ALICE	(1<<GFXB_AA_ALICE)
#define GFXB_AA_LISA		3
#define GFXF_AA_LISA		(1<<GFXB_AA_LISA)


/*
 *		HiRes(=Big) Agnus present; i.e. 
 *		1MB chipmem, big blits (none of interest so far) and programmable sync
 */
#define GFXG_OCS	(GFXF_HR_AGNUS)
/*
 *		HiRes Agnus/Denise present; we are running on ECS
 */
#define GFXG_ECS	(GFXF_HR_AGNUS|GFXF_HR_DENISE)
/*
 *		Alice and Lisa present; we are running on AGA
 */
#define GFXG_AGA	(GFXF_AA_ALICE|GFXF_AA_LISA)


#define SETCHIPREV_BEST	0xffffffff
#define HIRES				0x8000

struct View;


	/*
	 *		Graphics Library Functions
	 */

extern struct GfxBase *GfxBase;

static __inline void LoadView(struct View *view)
{
	register struct GfxBase* a6 __asm("a6") = GfxBase;
	register struct View *a1 __asm("a1") = view;

	__asm __volatile ("jsr (-0xde,a6)"
							: /* no output */
							: "r" (a6), "r" (a1)
							: "a0","a1","d0","d1", "memory");
}

static __inline u_long SetChipRev(unsigned long want)
{
	register u_long _res __asm("d0");
	register struct GfxBase *a6 __asm("a6") = GfxBase;
	register unsigned long d0 __asm("d0") = want;

	__asm __volatile ("jsr (-0x378,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (d0)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}


	/*
	 *		Intuition Library Definitions
	 */

struct NewScreen {
	short LeftEdge;
	short TopEdge;
	short Width;
	short Height;
	short Depth;
	u_char DetailPen;
	u_char BlockPen;
	u_short ViewModes;
	u_short Type;
	struct TextAttr *Font;
	u_char *DefaultTitle;
	struct Gadget *Gadgets;
	struct BitMap *CustomBitMap;
};

#define CUSTOMSCREEN				0x000f

struct NewWindow {
	short LeftEdge;
	short TopEdge;
	short Width;
	short Height;
	u_char DetailPen;
	u_char BlockPen;
	u_long IDCMPFlags;
	u_long Flags;
	struct Gadget *FirstGadget;
	struct Image *CheckMark;
	u_char *Title;
	struct Screen *Screen;
	struct BitMap *BitMap;
	short MinWidth;
	short MinHeight;
	u_short MaxWidth;
	u_short MaxHeight;
	u_short Type;
};

#define WFLG_SIMPLE_REFRESH	0x00000040
#define WFLG_ACTIVATE			0x00001000

struct Window;


	/*
	 *		Intuition Library Functions
	 */

extern struct IntuitionBase *IntuitionBase;

static __inline u_short CloseScreen(struct Screen *screen)
{
	register u_short _res __asm("d0");
	register struct IntuitionBase *a6 __asm("a6") = IntuitionBase;
	register struct Screen *a0 __asm("a0") = screen;

	__asm __volatile ("jsr (-0x42,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (a0)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline void CloseWindow(struct Window *window)
{
	register struct IntuitionBase *a6 __asm("a6") = IntuitionBase;
	register struct Window *a0 __asm("a0") = window;

	__asm __volatile ("jsr (-0x48,a6)"
							: /* no output */
							: "r" (a6), "r" (a0)
							: "a0","a1","d0","d1", "memory");
}

static __inline struct Screen *OpenScreen(struct NewScreen *newScreen)
{
	register struct Screen *_res __asm("d0");
	register struct IntuitionBase *a6 __asm("a6") = IntuitionBase;
	register struct NewScreen *a0 __asm("a0") = newScreen;

	__asm __volatile ("jsr (-0xc6,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (a0)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}

static __inline struct Window *OpenWindow(struct NewWindow *newWindow)
{
	register struct Window *_res __asm("d0");
	register struct IntuitionBase *a6 __asm("a6") = IntuitionBase;
	register struct NewWindow *a0 __asm("a0") = newWindow;

	__asm __volatile ("jsr (-0xcc,a6)"
							: "=r" (_res)
							: "r" (a6), "r" (a0)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}


	/*
	 *		Input Device Functions
	 */

extern struct InputBase *InputBase;

static __inline u_short PeekQualifier(void)
{
	register u_short _res __asm("d0");
	register struct InputBase *a6 __asm("a6") = InputBase;

	__asm __volatile ("jsr (-0x2a,a6)"
							: "=r" (_res)
							: "r" (a6)
							: "a0","a1","d0","d1", "memory");
	return(_res);
}


	/*
	 *		Bootstrap Support Functions
	 */

static __inline void change_stack(char *stackp)
{
	/*
	 *		Do ALWAYS compile this with optimization enabled!!
	 */

#ifndef __OPTIMIZE__
#error This program must be compiled with optimization enabled
#endif

	__asm __volatile ("movel %0,sp\n\t" :: "g" (stackp) : "sp");
}

static __inline void disable_mmu(void)
{
	if (SysBase->AttnFlags & AFF_68040)
		__asm __volatile ("moveq #0,d0;"
								".long 0x4e7b0003;"	/* movec d0,tc */
								".long 0x4e7b0004;"	/* movec d0,itt0 */
								".long 0x4e7b0005;"	/* movec d0,itt1 */
								".long 0x4e7b0006;"	/* movec d0,dtt0 */
								".long 0x4e7b0007"	/* movec d0,dtt1 */
								: /* no outputs */
								: /* no inputs */
								: "d0");
	else {
		__asm __volatile ("subl  #4,sp;"
								"pmove tc,sp@;"
								"bclr  #7,sp@;"
								"pmove sp@,tc;"
								"addl  #4,sp");
		if (SysBase->AttnFlags & AFF_68030)
			__asm __volatile ("clrl  sp@-;"
									".long 0xf0170800;" /* pmove sp@,tt0 */
									".long 0xf0170c00;" /* pmove sp@,tt1 */
									"addql #4,sp");
	}
}
